# -*- Mode: Perl -*-

###############################################################
#
#      File: LexWrapper.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: GPL
#
# Description: test Parse::Lex lexer
# 
###############################################################

package LexWrapper;
# System modules
use Carp;
require Exporter;
use Parse::Lex;
# 3rd party exstensions
# user extension modules
@ISA       = qw(Exporter);
@EXPORT    = qw();
@EXPORT_OK = qw();

###############################################################
# Package-global variables
###############################################################
our $DEBUG = 0;

###############################################################
# Constructor : new(%args)
#   + %args-keys:
#      lexer    : a Parse::Lex object
#      file     : a filehandle to analyze (defaults to UNDEF)
#      ignore   : a listref of token-NAME regexps to skip
#      stack    : a stack used by recurse(); array-ref of elts:
#                 { fh=>$fh,name=>$name,buf=>$buf,pos=>$pos,buflen=>$len }
#      _yylex   : lexing routine (default should be ok)
#      _yyerror : error reporing routine (default should be ok)
#      _yywarn  : warning routine (default should be ok)
#   + structure: hash
#     $lexer, $from, $ignore, $_yylex, $_yyerror, $_yywarn ]
###############################################################
sub new {
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {
		    lexer=>undef, # lexer
		    file=>undef, # file
		    filename=>'(none)',
		    ignore=>[],    # ignore
		    stack=>[],    # stack
		    _yylex=>undef, # _yylex
		    _yyerror=>undef, # _yyerror
		    _yywarn=>undef, # _yywarn
		   }, $class;
  # initialize
  return $self->initialize(@_);
}

sub DESTROY {
  my $self = shift;
  %$self = ();
};

###############################################################
# Accessors / manipulators
###############################################################
sub accessor {
  my ($key,$doupdate) = @_;
  return sub {
    my $self = shift;
    if (@_) { # set value
      $self->{$key} = shift;
      $self->_update() if ($doupdate);
    }
    return $self->{$key};
  };
}

*lexer = accessor('lexer',1);
*file = *from = accessor('file',1);
*filename = accessor('filename',1);
*ignore = accessor('ignore',1);
*stack = accessor('stack',0);

###############################################################
# You probably don't want to change these, anyways...
###############################################################
*_yylex = accessor('_yylex',0);
*_yyerror = accessor('_yyerror',0);
*_yywarn = accessor('_yywarn',0);

###############################################################
# Maintainence: _update([$lexer])
###############################################################
sub _update {
  my ($self,$lexer) = @_;
  $lexer = $self->lexer unless (defined($lexer));
  return undef unless (defined($lexer));
  $self->_yylex($self->_yylex_sub($lexer,$self->ignore));
  $self->_yyerror($self->_yyerror_sub($lexer,$self->filename));
  $self->_yywarn($self->_yywarn_sub($lexer,$self->filename));
  if (defined($self->{file})) {
    $self->lexer->from(ref($self->{file})
		       ? ${$self->{file}}
		       : $self->{file});
  } else {
    $self->lexer->from(''); # argh!
  }
}

###############################################################
# recurse($filehandle,[$filename]) => recursive lexing of $filehandle
###############################################################
sub recurse {
  my ($self,$fh,$filename) = @_;

  print
    ( __PACKAGE__ . "::recurse($fh,'$filename') called.\n")
      if ($DEBUG);

  return undef unless (defined($fh));
  $filename = filename_printable($fh) unless (defined($filename));

  my $current = {fh=>$self->file,
		 name=>$self->filename,
		 buf=>$self->{lexer}->buffer(),
		 buflen=>$self->{lexer}->length(),
		 pos=>$self->{lexer}->pos(),
		};
  push(@{$self->stack}, $current);

  print
    ("\trecurse(): pushing({",join(',',%$current),"})\n")
      if ($DEBUG);

  $self->lexer->reset();

  $self->file($fh);
  $self->filename($filename);

  $self->_update();
}

###############################################################
# yywrap():
#   + get the next token from nested input streams
###############################################################
sub yywrap {
  my $self = shift;

  print
    (__PACKAGE__ ,
     "::yywrap() called: stack=[",
     join(",",@{$self->stack}), "]\n")
      if ($DEBUG);

  return ('', undef) unless (@{$self->stack});
  my $popped = pop(@{$self->stack});

  print
    ("\t yywrap(): popped={", join(',',%$popped), "}\n") if ($DEBUG);

  $self->file($popped->{fh});
  $self->filename($popped->{name});
  $self->_update;
  $self->lexer->reset;
  $self->lexer->buffer($popped->{buf});
  $self->lexer->length($popped->{buflen});
  $self->lexer->pos($popped->{pos});

  return $self->yylex(@_);
}


###############################################################
# Initialization: initialize(%args)
###############################################################
sub initialize {
  my ($self, %args) = @_;
  my ($key,$val);
  while (($key,$val) = each(%args)) {
    if ($self->can($key)) {
      $self->$key($val);
    } else {
      warn("Unknown initizilation argument '$key' to object $self");
    }
  }
  return $self;
}

###############################################################
# _yylex_sub($Parse_Lex_Lexer,@skipnames) => 
#     anonymous sub ($MyParser) => (TOKENTYPE,TOKENVAL)
# - DESCRIPTION:
#   + Return a Yapp-friendly lexer routine using $Parse_Lex_Lexer
#   + @skipnames is a list of regexps which are matched against
#     the TYPES of incoming tokens; if a match occurs, the
#     matching token(s) are ignored by the lexer
# - REQUIREMENTS on returned sub:
#   + Yapp-compatible lexing routine
#   + read input stream and return token values to the parser
#   + our only argument ($MyParser) is the Parse::Yapp object itself
#   + We return a list ($TOKENTYPE, $TOKENVAL) of the next tokens
#     to the parser
#   + on end-of-input, we should return the list ('', undef)
# - CONVENTIONS for returned sub:
#   + use the following keys in $MyParser->YYData
#      Lexer : the Parse::Lex object for our lexer
#      --> OBSOLETE: now uses local closure-variable
#   + return token (EOF, '>>EOF<<') on first encountering
#     end-of-file, and only then start returning lack-of-input
#     tokens
###############################################################
sub _yylex_sub {
  my ($self,$lexer,$ignore) = @_;
  $lexer = $self->lexer unless (defined($lexer));
  $ignore = $self->ignore unless (defined($ignore));
  my ($token,$skipme);  # don't reallocate these, either
  return sub {
    if ($lexer->eoi) {   # *really* end-of-input
      return $self->yywrap();   # handle recursion
    }

  _YYL_TOKIGNORE:
    while ($token = $lexer->next()) {      # ignore these
      foreach $skipme (@$ignore) {
	next _YYL_TOKIGNORE if ($token->name =~ $skipme);
      }
      last;
    }
    return ('EOF', '>>EOF<<') if ($lexer->eoi);  # a single EOF token for the parser
    return ($token->name, $token->text) if (defined($token)); # usual case
  }
}

###############################################################
# _filename_printable($Parse_Lex_lexer,$filename)
# + returns a more or less printable version of $filename,
#   suitable for error and warning messages
###############################################################
sub _filename_printable {
  my ($self,$lexer,$filename) = @_;
  $lexer = $self->lexer unless (defined($lexer));
  $filename = $self->file unless (defined($filename));
  # HACK!
  my $oldwarn = $^W;
  $^W = 0;
  return $filename || undef unless (defined($lexer));
  if (!$filename && defined($self->from)) {
    return '(STDIN)' if (fileno($self->from) == 0);
    return $self->from;
  }
  $^W = $oldwarn;
  return $filename;
}

###############################################################
# _yyerror_sub($Parse_Lex_lexer,[$filename]) => 
#    anonymous sub ($parser) => undef
# + create generic error-reporting subs ('syntax error') for
#   Parse::Yapp parsers using Parse::Lex lexers
# + $filename is the filename which caused the error
###############################################################
sub _yyerror_sub {
  my ($self,$lexer,$filename) = @_;
  $lexer = $self->lexer unless (defined($lexer));
  $filename = $self->file unless (defined($filename));
  return undef unless (defined($lexer));
  $filename = _filename_printable(@_) unless (defined($filename));
  return sub {
    print STDERR 
      #"Error: $filename ($.): syntax error near token `",
      #$lexer->token->text, "'",
      #" of type ", $lexer->token->name, 
      #".\n";
      "Error: sytax error in file $filename at line $., near `",
      $lexer->token->text, "'.\n";
  }
}

###############################################################
# _yywarn_sub($Parse_Lex_lexer,$filename) =>
#    anonymous sub (@message) => undef
# + produces a warning sub suitable for use with Parse::Yapp
#   and Parse::Lex, which prints warnings of the form
#   $filename: $line: Warning: @messages\n
###############################################################
sub _yywarn_sub {
  my ($self,$lexer,$filename) = @_;
  $lexer = $self->lexer unless (defined($lexer));
  $filename = $self->file unless (defined($filename));
  return undef unless (defined($lexer));
  $filename = _filename_printable(@_) unless (defined($filename));
  return sub {
    #print STDERR "Warning: $filename ($.):", @_, "\n";
    print STDERR "Warning: ", @_, " in file $filename at line $..\n";
  }
}

###############################################################
# Top-level routines: 
#   + $myLexWrapper->yylex($parser) => (TOKNAME,TOKVAL)
#     --> probably better to do something like
#         $parser->YYParse(yylex => $myLexWrapper->_yylex,
#                          yyerror => $myLexWrapper->_yyerror);
#   + $myLexWrapper->yyerror($parser) => undef
#   + $myLexWrapper->yywarn(@message) => undef
#     --> this is actually useful -- i.e. if you set
#         a user-data variable in the parser
###############################################################
sub yylex { &{$_[0]->_yylex}(@_[1..$#_]); }
sub yyerror { &{$_[0]->_yyerror}(@_[1..$#_]); }
sub yywarn { &{$_[0]->_yywarn}(@_[1..$#_]); }


###############################################################
# dummylex: debug
###############################################################
sub dummylex {
  my $self = shift;
  my ($name,$text);
 TOKEN:while (1) {
    ($name,$text) = &{$self->_yylex}();
    if ($name) {
      print "Line $.\t";
      print "Type: ", $name, "\t";
      print "Content:->", $text, "<-\n";
      #if ($token->name =~ /STRING$/) {
      #eval 'package main; $stringval = ' . $token->text . ';';
      #print "\tSTRING-value(",$token->text,"):->$stringval<-";
      #}
      #print "\n";
      #print "$prompt" if ($token->name =~ '^NEWLINE');
    }
    else {
      last TOKEN;
    }
  }
}



###############################################################
# Autoload function passes unknown methods through
# to 'lexer' member
###############################################################
sub AUTOLOAD {
  my $self = shift;
  my $type = ref($self) || croak "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*:://; # strip package-part of name
  if (defined($self->lexer) && $self->lexer->can($name)) {
    return $self->lexer->$name(@_); # pass unknown methods to lexer
  } 
  croak("Unknown method '$name' for object $self");
}

1;
__END__
