###############################################################
#
#      File: SayWhat::OSC
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: OSC server for SayWhat generator
#
###############################################################

package SayWhat::OSC;

# system modules
use Carp;
use IO::File;

# user modules
use SayWhat::Base;
use OSC;
use OSC::Send;
use OSC::Dump;
use Socket;

@ISA = qw(SayWhat::Base);

our $VERSION = 0.01;

our $DEBUG = 0;

our $Prompt = 'OSCSayWhat> ';

our $DEFAULT_SYNTH_TYPE = __PACKAGE__->SUPER::DEFAULT_SYNTH_TYPE();
our %DEFAULT_SYNTH_ARGS =
  (
   SayWhat::Synthesizer::Festival => {
				      mode => 'wav',
				      samp_rate => 44100,
				     },
   SayWhat::Synthesizer::Rsynth => {
				    mode => 'wav',
				    samp_rate => 44100,
				   },
  );

sub DEFAULT_SYNTH_TYPE { return $DEFAULT_SYNTH_TYPE; }
sub DEFAULT_SYNTH_ARGS { return \%DEFAULT_SYNTH_ARGS; }

###############################################################
# Constructor: new(%args)
#  + %args keys
#    NEW
#     oschost=>$host : address on which to bind OSC::Dump socket
#     oscport=>$port : port on which to bind OSC::Dump socket
#     oscspeed=>$ms     : delay between token messages
#     oscspeedcoeff=>$ms : coefficient (* strlen($tok)) for osc msgs
#     oscbuflen=>$chars : buffer length for osc receive
#     oscmaxlen=>$chars : set maximum message-length in characters
#    INHERITED
#     wbound=>$str   : word-boundary string
#     grammar=>$g    : Grammar::ContextFree::Generator object
#     verbose=>$lvl  : verbosity level:
#                      -1 - silent
#                       0 - print only generated text
#                       1 - print status messages
#     synth=>$syn     : a SayWhat::Synthesizer object
#     synthtype=>$typ : type for synth
#     synthargs=>\%h  : args for new synth object (HASH-ref)
#  + internal structure:
#     utts     : information on generated utterances
#                { alias => { text=>$string, tree=>$tree } }
#     _uttidx  : internal utterance-counting index
###############################################################
sub new {
  my ($that,%args) = @_;
  my $sw = $that->SUPER::new(oscspeed=>10,
			     oscspeedcoeff=>100,
			     oscmaxlen=>1024,
			     oscbuflen=>1024,
			     %args);
  $sw->{oscdump} =
    OSC::Dump->new(LocalHost=>$sw->{oschost}||0,
		   LocalPort=>$sw->{oscport}||5151,
		  );
  if (!$sw->{oscdump}) {
    carp(__PACKAGE__ . "::new(): could not bind OSC socket: $!");
    return undef;
  }
  bless $sw, ref($that) || $that;
  return $sw;
}

###############################################################
# OSC server
###############################################################

# undef = $sw->server(\&callback)
#   + main server loop
#   + \&callback is called with a single-command string as an arg,
#     after setting up $sw->{oscsend} according to OSC msg address
sub server {
  my ($sw,$cb) = @_;
  my $buf = '';
  my $buflen = $sw->{oscbuflen}||1024;
  my ($parsed,$msg,
      $addr,$cmd,
      $shost,$sport);
  while ($sw->{peeraddr} = $sw->{oscdump}->recv($buf, $buflen, 0)) {
    if ($buf ne '') {
      @$sw{qw(peerport peerip)} = sockaddr_in($sw->{peeraddr});
      $sw->{peerhost} = inet_ntoa($sw->{peerip});

      $sw->verr(3,
		"$0: received packet from: ",
		"$sw->{peerhost}:$sw->{peerport}, ",
		" length: ", length($buf),
		"\n",
	       );

      $parsed = $sw->{oscdump}->parse($buf);
      foreach $msg (@$parsed) {
	($addr,$cmd) = @$msg{qw(addr vals)};
	$sw->verr(3, "$0: Parsed packet: addr=$addr cmd=$cmd\n");

	##-- addresses are of the form /HOST/PORT COMMAND ARGS...,
	$addr =~ s/^\/+//g;
	($shost,$sport) = split(/\/+/, $addr);
	$cmd =~ s/\,/ /g;

	##-- set up reply-socket
	if ($shost && $sport) {
	  $sw->{oscsendh} = OSC::Send->new(host=>$shost,port=>$sport);
	} else {
	  $sw->{oscsendh} = undef;
	}

	##-- run callback
	&$cb($cmd);
      }
    }
  }
}

###############################################################
# Prompting & Messages
###############################################################

##---------------------------------------------------------------
## $trimmed = $sw->trim_by_length(@strings)
##   + returns join(' ', @strings) trimmed to $sw->{oscmaxlen}
sub trim_by_length {
  my $sw = shift;
  my $str = join(' ',@_);
  if (defined($sw->{oscmaxlen}) && length($str) > $sw->{oscmaxlen}) {
    substr($str,$sw->{oscmaxlen}-3) = '...' ;
  }
  return $str;
}


##---------------------------------------------------------------
## @osc_token_messages = $sw->osc_tokens(@untokenized_strings)
sub osc_tokens {
  my ($sw) = shift;
  my ($msg);
  return
    (map {
      $msg = $_;
      $msg =~ s/[\;\,]/ /g;
      (map {"/saywhat/t,$_"} split(/\s+/, $msg))
    } @_);
}

##---------------------------------------------------------------
## $sw->sendosc(@messages)
##  + send each elt of @messages in turn to our OSC peer
##    (indicated by $sw->{oscsendh})
sub sendosc {
  my $sw = shift;
  $sw->verr(5, __PACKAGE__ , "::sendosc() called: ", join(';',@_), "\n");
  return if (!$sw->{oscsendh});
  my ($args);
  foreach (@_) {
    select(undef,undef,undef,$sw->{oscspeed}/1000)
       if (defined($sw->{oscspeed}));
    $sw->{oscsendh}->send($_);
    if (defined($sw->{oscspeedcoeff})) {
       ($args=$_) =~ s/^[^,]*//;
       select(undef,undef,undef,length($args)*$sw->{oscspeedcoeff}/1000)
    }
  }
}

###############################################################
# Status
###############################################################

##---------------------------------------------------------------
## $rv = $sw->status($op,$rv,@args)
sub status {
  my ($sw,$op,$rv) = (shift,shift,shift);
  $sw->SUPER::status($op,$rv,@_) if ($sw->{verbose} >= 3);
  if ($op eq 'generate') {
    my ($sw2,$id,$str,$tree) = @_;
    $sw->sendosc("/saywhat/id,$id",
		 $sw->osc_tokens($sw->trim_by_length($str)),
		 '/saywhat/EOU,1',
		);
  } else {
    $sw->sendosc("/saywhat/$op,"
		 . join(',',$rv,
			(map { defined($_) ? $_ : 'undef' } @_))
		);
  }
  return $rv;
}


1;
__END__

