###############################################################
#
#      File: SayWhat::Synthesizer.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: Abstract class for SayWhat speech synthesizers
#
###############################################################

package SayWhat::Synthesizer;
use Exporter;
#use File::Temp qw(tmpfile tmpdir);
use Carp;

our $VERSION = 0.01;

@ISA = qw(Exporter);

our %EXPORT_TAGS = (
		    rc => [qw($RC_FAILED $RC_OK)],
		    cmds => [qw(%PLAY_CMDS)],
		    gensym => [qw(&gensym_utt &_utt_gensym)],
		   );
$EXPORT_TAGS{all} = [map { @$_ } values(%EXPORT_TAGS)];

our @EXPORT_OK   = @{$EXPORT_TAGS{all}};

our @EXPORT      = qw();

###############################################################
# Globals
###############################################################

our %PLAY_CMDS = (
		  raw => 'play --rate=$syn->{samp_rate} --channels=1 --type=sw',
		  au => 'play',
		  wav => 'play',
		 );

our ($RC_FAILED,$RC_OK) = (0,1);

###############################################################
# Constructor: $syn = SayWhat::Synthesizer->new(%args)
# + %args keys
#   samp_rate => $rate,     # sampling rate for output files
#   type => $type,          # type of synthesizer (class name or suffix)
# + internal state:
#   _uttidx => $i,          # index for '_utt_gensym()'
###############################################################
sub new {
  my ($that,%args) = @_;
  my $class = ref($that) || $that;

  my $type = $args{type};
  delete($args{type});

  my ($self);
  if (defined($type)) {
    ## canonicalize type-name
    $type =  "$class::$type" unless UNIVERSAL::isa($type, $class);
    unless (UNIVERSAL::isa($type, $class)) {
      carp("warning: specified synthesier type '$type' is not derived from '$class'\n");
    }
    $self = $type->new(%args);
  } else {
    $self = bless {}, $class;
  }
  $self->configure(%args);

  return $self;
}

###############################################################
# accessors
###############################################################

# $rate = $syn->samp_rate(); $syn->samp_rate($rate);
sub samp_rate {
  if ($#_ > 0) {
    return $_[0]->{samp_rate} = $_[1];
  }
  return $_[0]->{samp_rate};
}

# %config = $syn->configure();
# $val = $syn->configure($key);
# $syn = $syn->configure(%config);
sub configure {
  my $syn = shift;
  if (scalar(@_) == 0) {
    # configure() : get whole config
    return %$syn;
  } elsif (scalar(@_) == 1) {
    # configure($key) : get value
    return $syn->{$_[0]};
  } else {
    # configure(%args) : set config
    %$syn = (%$syn,@_);
    return $syn;
  }
}

###############################################################
# error-reporting
###############################################################

# $rc = $syn->dummy('name');
sub dummy {
  carp("warning: method ", __PACKAGE__, "::$_[1] not defined!");
  return $RC_FAILED;
}

###############################################################
# subs: utterance maintainence
#   + $rc = gensym_utt()
#   + $rc = add_utterance($id,$text);
#   + $rc = del_utterance($id);
#   + $rc = rename_utterance($id);
#   + $rc = synth_utterance($id);
#   + $rc = say_utterance($id);
#   + $rc = save_utterance($id,[$file,[$mode]]);
###############################################################
sub gensym_utt {
  my $self = shift;
  my $sym = "utt" . (defined($self->{_uttidx}) ? $self->{_uttidx} : 0);
  $self->{_uttidx}++;
  return $sym;
}
*_utt_gensym = \&gensym_utt;

sub add_utterance { $_[0]->dummy('add_utterance'); }
sub del_utterance { $_[0]->dummy('del_utterance'); }
sub synth_utterance { $_[0]->dummy('synth_utterance'); }
sub say_utterance { $_[0]->dummy('say_utterance'); }
sub save_utterance { $_[0]->dummy('save_utterance'); }

###############################################################
# subs: start/stop
###############################################################

# $syn->start();
#  + starts synth
sub start { $_[0]->dummy('start'); }

# $syn->stop();
#  + stops synth
sub stop { $_[0]->dummy('stop'); }


##-------------------------------------
## Dummy class
package SayWhat::Synthesizer::Dummy;
@ISA = qw(SayWhat::Synthesizer);

our $VERBOSE = 0;

# $rc = $syn->dummy('name');
sub dummy {
  my ($sw,$name) = (shift,shift);
  print STDERR (__PACKAGE__, "::$name(", join(',',@_), ")\n") if ($VERBOSE);
  return $RC_OK;
}

1;

__END__
