###############################################################
#
#      File: SayWhat::Synthesizer::Festival.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: SayWhat speech synthesizer interface to Festival
#
###############################################################

package SayWhat::Synthesizer::Festival;

# system modules
use Exporter;
use Carp;

# user modules
use SayWhat::Synthesizer qw(:rc :cmds :gensym);

# 3rd-party modules
use Speech::Festival;

our $VERSION = 0.02;

@ISA = qw(SayWhat::Synthesizer);

###############################################################
# Globals
###############################################################

# %DEFAULTS : default values for new() keywords
our %DEFAULTS = (
		 mode => 'wav',
		 #smode => 'Text',
		 #samp_rate => 16000,
		 #voice => 'us1_mbrola',
		 #style => 'default',
		 autogc => 1,
		 host => 'localhost',
		 port => 1314,
		 timeout => 0, # timeout (secs); 0 = no timeout; undef = don't wait for result
		 verbose => 2, # verbose: 0:errors, 1:normal; 1:festival-requests
		);

###############################################################
# Constructor: $syn = SayWhat::Synthesizer::Rsynth->new(%args)
# + %args keys:
#   ## INHERITED
#   mode => $mode,          # output mode ('au'|'raw'|'wav') ; default='wav'
#   samp_rate => $rate,     # sampling rate for output files
#   passthru => \@args,     # args for festival
#   dummy => $bool,         # dummy args
#   ## NEW
#   host => $host,          # festival-server host
#   port => $port,          # festival-server port
#   voice => $voice,        # festival voice
#   style => $style,        # festival style
#   timeout => $secs,       # festival result timeout
#   autogc=>$bool,          # whether to garbage-collect automatically
# + internal structure
#   utts => {               # text to scheme-varname mapping
#            $id => { name=>$name, text=>$text, synthed=>$bool },
#           },
#   festival => $festival,  # a Speech::Festival object
#   running => $bool,
###############################################################
sub new {
  my ($that,%args) = @_;
  my $class = ref($that) || $that;
  my $self = bless {
		    %DEFAULTS,
		    %args,
		   }, $class;

  ##-- dummy-mode
  if ($args{dummy}) {
    $self->{festival} = FileHandle->new(">&STDOUT");
    die("could not dup to stdout for dummy festival filehandle: $!")
      unless (defined($self->{festival}));
    $self->{festival}->autoflush(1);
  }
  delete($self{dummy});

  ## real-mode: create festival object (but don't connect yet)
  if (!defined($self->{festival})) {
    $self->{festival} = Speech::Festival->new($self->{host},$self->{port});
  }
  #if (!$self->{festival}) {
  #  carp("$class->new(): could not create Speech::Festival object!\n");
  #}
  #else {
  #  $self->voice($self->{voice}) if (defined($self->{voice}));
  #  $self->style($self->{style}) if (defined($self->{style}));
  #}
  #delete(@{$self}{qw(voice style)});

  return $self;
}

###############################################################
# accessors / configuration
###############################################################

# %config = $syn->configure();
# $val = $syn->configure($key);
# $syn = $syn->configure(%config);
sub configure {
  my $syn = shift;
  return $syn->{$_[0]} if (scalar(@_) == 1); # get value

  if (scalar(@_) == 0) {
    # configure() : get whole config
    my %config = (%$syn,
		  host=>$syn->{host},
		  port=>$syn->{port},
		  style=>$syn->style()||'undef',
		  voice=>$syn->voice()||'undef',
		  autogc=>$syn->{autogc}||0,
		 );
    delete(@config{qw(festival utts running _uttidx)});
    return %config;
  } else {
    # configure(%args) : set config
    my ($k,$v);
    while (($k,$v) = (splice(@_,0,2))) {
      ## -- ignore these
      next if ($k eq 'dummy');
      $v = undef if ($v eq 'undef');
      ## -- these *would* require special handling
      ##    ... but they make request() hang later!
      ##    ... this seems to be a festival bug (happens w/ 'festival_client --server carrot', too!)
      if    ($k eq 'host' && $v ne $syn->{host}) {
	$syn->stop;
	$syn->{host} = $v;
	#$syn->reinitialize_utts();
      }
      elsif ($k eq 'port' && $v ne $syn->{port}) {
	$syn->stop;
	$syn->{port} = $v;
	#$syn->reinitialize_utts();
      }
      ## -- these *do* require special handling (voice,style,gc)
      elsif (my $code = $syn->can($k)) { &$code($syn,$v); }
      else {
	# normal case
	$syn->{$k} = $v;
      }
    }
    return $syn;
  }
}


###############################################################
# handy
###############################################################
our %MODE2TYPE =
  (
   raw => 'raw',
   wav => 'riff',
   au => 'au',
  );

###############################################################
# subs: utterance maintainence
#   + $rc = add_utterance($id,$text);
#   + $rc = del_utterance($id);
#   + $rc = synth_utterance($id);
#   + $rc = say_utterance($id);
#   + $rc = save_utterance($id,[$file,[$mode]]);
###############################################################

###############################################################
# add/delete/rename
sub add_utterance {
  my ($syn,$id,$text) = @_;
  #my $scheme_id = $syn->_utt_gensym();
  my $scheme_id = $id;
  $id = $scheme_id unless (defined($id));
  $syn->{utts}{$id} = {name=>"$scheme_id",text=>$text};
  return $syn->send_festival("(set! $scheme_id (Utterance Text \""
			     .(defined($text) ? $text : '')."\"))");
}
sub rename_utterance {
  my ($syn,$oldid,$newid) = @_;
  return $RC_OK if ($oldid eq $newid);
  unless ($syn->add_utterance($newid,$syn->{utts}{$newid}{text}) eq $RC_OK
	  && $syn->del_utterance($oldid) eq $RC_OK)
    {
      carp(ref($syn)."->rename_utterance() failed!");
      return $RC_FAILED;
    }
  delete $syn->{utts}{$oldid};
  return $RC_OK;
}
sub del_utterance {
  my ($syn,$id) = @_;
  my $utt = $syn->{utts}{$id};
  if (!defined($utt)) {
    carp(ref($syn)."->del_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  return $RC_FAILED unless
    ($syn->send_festival("(set! $utt->{name} nil)") eq $RC_OK
     && $syn->send_festival('(gc)') eq $RC_OK);
  delete($syn->{utts}{$id});
  return $RC_OK;
}

###############################################################
# synth/say/save
sub synth_utterance {
  my ($syn,$id) = @_;
  if (!defined($syn->{utts}{$id})) {
    carp(ref($syn)."->synth_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  return $RC_FAILED
    unless ($syn->send_festival("(utt.synth $syn->{utts}{$id}{name})") eq $RC_OK);
  $syn->{utts}{$id}{synthed} = 1;
  return $RC_OK;
}
sub say_utterance {
  my ($syn,$id) = @_;
  if (!defined($syn->{utts}{$id})) {
    carp(ref($syn)."->say_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  return $syn->send_festival("(utt.play $syn->{utts}{$id}{name})");
}
sub save_utterance {
  my ($syn,$id,$file,$mode) = @_;
  ## set up mode & filename defaults
  $mode = $syn->{mode} unless (defined($mode));
  $file = "$id.$mode" unless (defined($file));

  ## check for bad id
  my $utt = $syn->{utts}{$id};
  if (!defined($utt)) {
    carp(ref($syn)."->save_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }

  ## ensure that it's been synthesized
  unless ($utt->{synthed}) {
    return $RC_FAILED unless ($syn->synth_utterance($id) eq $RC_OK);
  }
  ## resample it
  return $RC_FAILED
    unless ($syn->send_festival("(utt.wave.resample $utt->{name} $syn->{samp_rate})") eq $RC_OK);
  ## save it
  my $filetype = $MODE2TYPE{$mode} || $mode;
  my $rc = $syn->send_festival("(utt.save.wave $utt->{name} \"$file\" \'$filetype)");
  return ($rc,$file,$mode);
}


###############################################################
# festival: connect/disconnect
###############################################################

## $bool = $syn->connect_festival();
##    + sets $syn->{festival} & connects
sub connect_festival {
  my $syn = shift;
  $syn->{festival} = Speech::Festival->new($self->{host},$self->{port})
    if (!defined($syn->{festival}));
  return $syn->{festival} ? $syn->{festival}->conn() : undef;
}

## $bool = $syn->disconnect_festival();
##    + disconnects from festival server
sub disconnect_festival {
  my $syn = shift;
  return $syn->{festival} ? $syn->{festival}->disconnect() : undef;
}

###############################################################
# subs: maintainence
###############################################################

# $syn->start();
#  + starts synth
sub start {
  my $syn = shift;
  unless ($syn->connect_festival()) {
    carp(ref($syn)."->start(): error: could not connect to festival server");
    return $RC_FAILED;
  }
  $syn->voice($self->{voice}) if (defined($syn->{voice}));
  $syn->style($self->{style}) if (defined($syn->{style}));
  $syn->{running} = 1;
  return $RC_OK;
}

# $syn->stop();
#  + stops synth
sub stop {
  my $syn = shift;
  ## remove utterances
  #%{$_[0]->{utts}} = ();
  foreach my $u (values(%{$syn->{utts}})) {
    $u->{synthed} = 0;
  }
  ## close down synth
  $syn->{running} = 0;
  return $syn->disconnect_festival;
}

# $rc = $syn->reinitialize_utts();
#   + re-add()'s all utterance on synth re-connect
sub reinitialize_utts {
  my $syn = shift;
  my $utts = $syn->{utts};
  my ($k,$utt);
  while (($k,$utt) = each(%$utts)) {
    unless ($syn->send_festival("(set! $utt->{name} (Utterance Text \"$utt->{text}\"))") eq $RC_OK) {
      carp(ref($syn)." Error: utterance re-initialization failed!");
      return $RC_FAILED;
    }
    $utt->{synthed} = 0;
  }
  return $RC_OK;
}


# OTHER STUFF

###############################################################
# voice / style
###############################################################

*set_voice = \&voice;
*get_voice = \&voice;
*set_style = \&style;
*get_style = \&style;

# $voice = $syn->voice();
# $voice = $syn->voice($voice)
# $style = $syn->style();
# $style = $syn->style($style)
sub voice {
  my $syn = shift;
  if (@_) {
    # set voice
    $syn->send_festival("(voice.select \'$_[0])");
  }
  # get voice
  my $voice = ($syn->send_festival("current-voice"))[1];
  return undef unless (defined($voice));
  chomp($voice);
  return $voice;
}

sub style {
  my $syn = shift;
  if (@_) {
    # set voice
    $syn->send_festival("(Style \'$_[0])");
  }
  # get voice
  my $style = ($syn->send_festival("current_style"))[1];
  return undef unless (defined($style));
  chomp($style);
  return $style;
}

sub gc {
  my $syn = shift;
  return $syn->{festival} ? $syn->send_festival('(gc)') : $RC_OK;
}

###############################################################
# send_festival()
#  LIST CONTEXT:
#   ($type,$data) = send_festival($scheme_form)
#   + does what you'd expect, returns UNDEF if result timed out
#  SCALAR CONTEXT:
#   $rc = send_festival($scheme_form)
###############################################################
sub send_festival {
  my $self = shift;
  my $fest = $self->{festival};

  return undef unless ($self->{running} || $self->start eq $RC_OK);

  if (defined($fest)) {
    if (ref($fest) =~ /^Speech::Festival/) {
      ##-- real festival client mode
      my $req = shift;
      $fest->request($req);
      print(STDERR $req, "\n") if ($self->{verbose} > 1);

      if (defined($self->{timeout})) {
	##-- timeout is defined -- try to get results
	my @waitargs = qw();

	##-- allow real timeout only if {timeout} is nonzero
	@waitargs = ($self->{timeout}) if ($self->{timeout} != 0);
	my @result = qw();

	##-- get two result-sets [1st is request result; 2nd is status (?)]
	if ($fest->result_waiting || $fest->wait_for_result(@waitargs)) {
	  push(@result,$fest->get_result());
	}
	if ($result[0] ne $Speech::Festival::ERROR &&
	    ($fest->result_waiting || $fest->wait_for_result(@waitargs))) {
	  push(@result,$fest->get_result());
	}
	return @result if (wantarray);
	return $result[0] eq $Speech::Festival::ERROR ? $RC_FAILED : $RC_OK;
      }
      ## timeout was undefined or no result was waiting
      return undef;
    }
    ## dummy mode (print to filehandle)
    return map { print $fest $_,"\n" } @_;
  }
  ## silent mode
  return undef;
}
